=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed April / June 2008 by Fredo6

# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted, provided that the above
# copyright notice appear in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   LineOnSurface.rb
# Original Date	:   14 May 2008 - version 1.1
# Revisions		:	04 Jun 2008 - version 1.2
# Type			:   Sketchup Tools
# Description	:   Suite of Tools to draw on a surface
# Menu Items	:   Tools --> "Line on Surface"
# Toolbar		:   Name = "Surface Operations"
# Context Menu	:   All options
# Usage			:   See Tutorial and Quick Ref Card in PDF format
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

require 'sketchup.rb'
require 'LibTraductor.rb'			# for language translation
require 'LibOnSurface.rb'		# for basic routines

module SUToolsOnSurface

#Constants for LineOnSurface Module (do not translate)	
DOS_ICON_LINE = "Line"
DOS_ICON_CLINE = "Cline"
DOS_CURSOR_LINE = "Line"
DOS_CURSOR_CLINE = "Cline"

STR_Line_Title = ["Draw Line on Surface",
	              "|FR| Trac\ de ligne sur une surface"]					 
MSG_Line_Origin = ["Enter Origin",
                   "|FR| Point d'origine"]
MSG_Line_End = ["Enter End Point",
                "|FR| Point de Fin"]
MSG_Line_Execution = ["Drawing line on Surface",
                      "|FR| Trac\ de la ligne sur la surface"]

				 
#--------------------------------------------------------------------------------------------------------------
# Top Calling functions: create the classes and launch the tools
#--------------------------------------------------------------------------------------------------------------			 				   

def SUToolsOnSurface.launch_line
	@tool_line = TOSToolLine.new true unless @tool_line
	Sketchup.active_model.select_tool @tool_line
end

def SUToolsOnSurface.launch_cline
	@tool_cline = TOSToolLine.new false unless @tool_cline
	Sketchup.active_model.select_tool @tool_cline
end

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# TOSToolLine: Tool to draw line (plain or construction) on a surface
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
					
class TOSToolLine

def initialize(linemode)
	@linemode = linemode

	#Loading strings and cursors
	Traductor.load_translation SUToolsOnSurface, /MSG_/, binding, "@msg_"
	@idcursor_line = OFSG.create_cursor DOS_CURSOR_LINE, 3, 31
	@idcursor_cline = OFSG.create_cursor DOS_CURSOR_CLINE, 3, 31

	#initializing variables
	@ip_origin = Sketchup::InputPoint.new
	@ip_end = Sketchup::InputPoint.new
	@ip = Sketchup::InputPoint.new
	
	#Initializing parameters
	@ofsgrp = CommonGroup.create
	@option_group = OFS_DEFAULT_Group
	@option_cpoint = (linemode) ? OFS_DEFAULT_CPoint_L : OFS_DEFAULT_CPoint_C
	@option_nocurves = ! OFS_DEFAULT_GenCurve
	@prev_dist = 0
	@prev_dist0 = 0
	@prev_vec = nil
	@prev_vec0 = nil
	
	#Group handling
	
end

def activate
	@model = Sketchup.active_model
	@selection = @model.selection
	@entities = @model.active_entities
	
	@option_group = @ofsgrp.get_option_group	
	
	@pt_origin = nil
	@pt_end = nil
	@state = STATE_ORIGIN
	@enter_down = false
	@pts = []
	@distance = 0
	@parcours = []
	@edges = nil
	@moved = false
	
	info_show
end

def deactivate(view)
	view.invalidate
end

def onCancel(flag, view)
	#User did an Undo
	case flag
	when 1, 2	#Undo or reselect the tool
		activate
		return
	when 0	#user pressed Escape
		return  if (@state == STATE_ORIGIN)  #Exiting the tool
		set_state @state - 1
	end
end

def onSetCursor
	UI::set_cursor (@linemode) ? @idcursor_line : @idcursor_cline
end

#Compute the path of the line, both for interactive session and final construction
def compute_path(store=false)
	return @pts = [] unless @pt_origin && @pt_end
	parcours = Junction.calculate @mark_origin, @mark_end
	@pts = []
	parcours.each { |mk| @pts.push mk.pt }
	compute_distance
	if store
		@list_coseg = []
		OFSG.compute_coseg(parcours, @list_coseg) if @linemode
		@parcours = parcours 
	end	
	return @pts
end
	
#Computing distance
def compute_distance
	return 0 if @pts.length < 2
	nb = @pts.length - 2
	@distance = 0.0.to_l
	for i in 0..nb
		@distance += @pts[i].distance @pts[i+1]
	end
end

#Finalization of the line drawing
def compute_junction
	#calculating the parcours
	compute_path true
	
	#Drawing the parcours
	execute_drawing

	#Chaining with next point
	prepare_chaining
end

#Used for Redo with imposed distance
def execute_to_distance(mark, vecdir, len)
	@parcours = Junction.to_distance mark, vecdir, len
	@pts = []
	@parcours.each { |mk| @pts.push mk.pt }

	#Drawing the parcours
	execute_drawing	
end

#Creation method for the line
def execute_drawing
	return if @pts.length < 2
	
	#Storing distance for further Redo
	push_previous @pts

	@model.start_operation @msg_Line_Execution

	#identifying the Group if needed
	if @option_group
		grp = @ofsgrp.get_current
		unless grp
			@model.abort_operation
			return
		end	
		entities = grp.entities
	else
		entities = @entities
	end
	
	#creating the new edges or construction lines
	attr = '-'
	if @linemode
		@edges = []
		OFSG.commit_line(entities, @pts, attr, @option_nocurves, @list_coseg, @edges, [@pts.first, @pts.last])
	else
		nb = @pts.length - 2
		for i in 0..nb
			cline = entities.add_cline @pts[i], @pts[i+1]
			OFSG.set_cline_attribute cline, attr
		end
	end	
	if @option_cpoint
		@pts.each do |pt|
			cpoint = entities.add_cpoint pt 
			OFSG.set_cline_attribute cpoint, attr
		end
	end	
	@model.commit_operation
	
end

#Perfrom correction of distance after operation
def correct_previous(len)
	#Checking if this is applicable
	return false unless @edges
	@edges.each { |e| return false unless e.valid? }
	return false unless @parcours && @parcours.length >= 2
	
	#Computing the origin
	mark = @parcours[0]
	vecdir = mark.pt.vector_to @parcours[1].pt
		
	Sketchup.undo
	execute_to_distance mark, vecdir, len
	prepare_chaining
	true
end

def prepare_chaining(view=nil)
	return unless @parcours.length > 0
	view = @model.active_view unless view
	@mark_origin = @parcours.last
	@pt_origin = @mark_origin.pt
	@face_origin = @mark_origin.face
	@mark_end = @parcours.last
	ptv = view.screen_coords @pt_origin
	@xorig = ptv.x
	@yorig = ptv.y
	@ip_origin.pick view, @xorig, @yorig
	set_state STATE_END
end

#Control the states of the tool
def set_state(state)
	@state = state
	if state == STATE_EXECUTION 
		compute_junction
	elsif state <= STATE_END
		@moved = false
	end
	info_show
end

def getMenu(menu)
	if (@state >= STATE_END)
		menu.add_item(@msg_MnuDone) { compute_junction }
		menu.add_separator
	end	
	if (@prev_dist > 0)
		tx = Traductor[MSG_MnuRedo, @prev_dist.to_l]
		menu.add_item(tx) { menu_redo }
		menu.add_separator
	end
	option_contextual_menu menu
	true
end

#Populate the options in the Contextual menu
def option_contextual_menu(menu)	
	txcur = @msg_MnuCurrent
	text = @msg_MnuGroup + " #{txcur} " + Traductor[DLG_EnumYesNo[(@option_group) ? 'Y' : 'N']] + 
	       ") --> " + OFS___Group
	menu.add_item(text) { self.change_option_group }
	text = @msg_MnuNoCurves + " #{txcur} " + Traductor[DLG_EnumYesNo[(@option_nocurves) ? 'N' : 'Y']] + 
		   ") --> " + OFS___NoCurves
	menu.add_item(text) { self.change_option_nocurves }
	text = @msg_MnuCPoint + " #{txcur} " + Traductor[DLG_EnumYesNo[(@option_cpoint) ? 'Y' : 'N']] + 
	       ") --> " + OFS___CPoint
	menu.add_item(text) { self.change_option_cpoint }
	text = @msg_MnuLineMode + " #{txcur} " + Traductor[(@linemode) ? STR_ModeLine : STR_ModeCLine]  + 
	       ") --> " + OFS___LineMode
	menu.add_item(text) { self.change_option_linemode }
end

def change_option_linemode
	@linemode = !@linemode
	onSetCursor
end

def change_option_group
	@option_group = @ofsgrp.set_option_group(!@option_group)
end

def change_option_cpoint
	@option_cpoint = !@option_cpoint
end
	
def change_option_nocurves
	@option_nocurves = !@option_nocurves
end

def onLButtonDown(flags, x, y, view)
	@time_mouse_down = Time.now
	@xdown = x
    @ydown = y
	if (@state == STATE_END)
		return if (@xdown - @xorig).abs < 2 && (@ydown - @yorig).abs < 2
	end	
	set_state @state + 1
end

#Double Click to repeat with same length
def onLButtonDoubleClick(flags, x, y, view)
	redo_previous
end

def push_previous(pts)
	nb = pts.length - 2
	d = 0
	if (nb >= 0)
		d = 0
		for i in 0..nb
			d += pts[i].distance pts[i+1]
		end	
	end	
	return if d == 0
	
	#swapping the stored values	
	@prev_dist = @prev_dist0
	@prev_vec = (@prev_vec0) ? @prev_vec0.clone : nil
	@prev_dist0 = d
	@prev_vec0 = pts[0].vector_to pts[1]
end

#Correction of length after operation
def redo_previous
	if (@pt_end && @prev_dist > 0 && @moved == false)
		correct_previous @prev_dist
	elsif (@moved == false && @prev_vec0 && @prev_dist0 > 0)
		mark = OFSG.create_mark @pt_origin, @face_origin
		execute_to_distance(mark, @prev_vec0, @prev_dist0)
		prepare_chaining
	else
		UI.beep
	end	
end

#Redo, when called from contextual menu
def menu_redo
	if (@pt_end && @prev_dist0 > 0 && @moved)
		mark = OFSG.create_mark @pt_origin, @face_origin
		vecdir = @pts[0].vector_to @pts[1]
		execute_to_distance(mark, vecdir, @prev_dist0)
		prepare_chaining
	elsif (@moved == false && @prev_vec0 && @prev_dist0 > 0)
		mark = OFSG.create_mark @pt_origin, @face_origin
		execute_to_distance(mark, @prev_vec0, @prev_dist0)
		prepare_chaining
	else
		UI.beep
	end	
end

#Key Up
def onKeyUp(key, rpt, flags, view)
	key = Traductor.check_key key, flags, true

	case key
		#Toggling between fixed and variable length
		when COPY_MODIFIER_KEY
			change_option_linemode if @control_down
			onMouseMove(flags, @xend, @yend, view) if (@state >= STATE_END)
			view.invalidate
			info_show
	end	
	@control_down = false
end

#Key down
def onKeyDown(key, rpt, flags, view)
	key = Traductor.check_key key, flags, false

	case key			
		#Calling options
		when COPY_MODIFIER_KEY
			@control_down = true
			return
		when TABLE_FKEY[OFS___Group]
			change_option_group
		when TABLE_FKEY[OFS___NoCurves]
			change_option_nocurves
		when TABLE_FKEY[OFS___LineMode] 
			change_option_linemode
		when TABLE_FKEY[OFS___CPoint]
			change_option_cpoint
			
		when 13
			@enter_down = true
			@control_down = false
			return
		else
			@control_down = false
			return
			
	end	
	@control_down = false
	
	onMouseMove(flags, @xend, @yend, view) if (@state >= STATE_END)
	view.invalidate
	info_show
end

#Input of length in the VCB
def onUserText(text, view) 
	@enter_down = false
	begin
		len = text.to_l
		if len == 0
			UI.beep
			return				
		end	
		
		#Modify the distance for the previously entered line
		if @state == STATE_ORIGIN || @pt_end == nil || @pt_end == @pt_origin || @distance == 0
			UI.beep unless correct_previous len
			return
		elsif @pt_end
			vecdir = @pt_origin.vector_to @pt_end
			if vecdir.length > 0.0
				mark = OFSG.create_mark @pt_origin, @face_origin
				execute_to_distance mark, vecdir, len
				prepare_chaining view
			else
				UI.beep
			end	
		end	
	rescue
		UI.beep
	end	
end

#Mouse Move method
def onMouseMove(flags, x, y, view)
	#view.lock_inference
	#Origin Point
    if (@state == STATE_ORIGIN)
		@xorig = x
		@yorig = y
        @ip.pick view, x, y
        
		@ip_origin.copy! @ip
		@mark_origin = OFSG.mark_from_inputpoint view, @ip_origin, x, y
		@pt_origin = @mark_origin.pt
		@face_origin = @mark_origin.face
		
		@pt_end = nil
		@moved = false
		
	#End Point	
	elsif (@state == STATE_END)
		@xend = x
		@yend = y		
        @ip.pick view, x, y
        if (@ip != @ip_origin) && @ip.valid?
            @ip_end.copy! @ip
			@mark_end = OFSG.mark_from_inputpoint view, @ip_end, x, y
			@pt_end = @mark_end.pt
			@face_end = @mark_end.face
			@moved = true
        end
	
	end
	tt = @ip.tooltip
	view.tooltip = tt
	view.invalidate
	info_show
end	

#Draw method for tool
def draw(view)
	if (@state >= STATE_ORIGIN)
	    if @ip_origin.valid? && @ip_origin.display?
			if (@face_origin == nil)
				view.draw_points @pt_origin, 15, 3, 'purple'
			elsif @pt_origin && @face_origin.classify_point(@pt_origin) >= 8
				view.draw_points @pt_origin, 15, 5, 'orange'
			else	
				@ip_origin.draw view
				view.draw_points @pt_origin, 15, 3, 'purple' if @pt_origin && @pt_origin != @ip_origin.position
			end	
        end
	end	
	
	if (@state >= STATE_END)
	    if @ip_end.valid? && @ip_end.display? && @pt_end
			if (@face_end == nil)
				view.draw_points @pt_end, 15, 3, 'purple'
			elsif @face_end.classify_point(@pt_end) >= 8 
				view.draw_points @pt_end, 15, 5, 'orange'
			else	
				@ip_end.draw view
			end	
        end
		pts = compute_path
		if @linemode
			if (@option_group)
				view.drawing_color = 'darkred'
				view.line_stipple = "-.-"
				view.line_width = 2
			else	
				view.drawing_color = 'black' 
				view.line_stipple = ""
				view.line_width = 1
			end	
		else
			view.drawing_color = (@option_group) ? 'orange' : 'black'
			view.line_stipple = "_"
			view.line_width = 1
		end
		view.draw GL_LINE_STRIP, pts if pts.length > 1
		view.line_stipple = ""
		view.draw_points pts[1..-2], 10, 3, 'black' if @option_cpoint && pts.length > 2
	end	
	#OFSG.watchlist_draw view
end

#display information in the Sketchup status bar
def info_show
	msg = "[" + Traductor[(@linemode) ? STR_ModeLine : STR_ModeCLine] + "] "

	case @state
	when STATE_ORIGIN
		msg += @msg_Line_Origin
	when STATE_END
		msg += @msg_Line_End
	when STATE_EXECUTION
		msg += @msg_Line_Execution
	end
	
	compute_distance
	d = @distance

	msg += " [" + @msg_Group + "]" if (@option_group)
	msg += " [" + @msg_NoCurves + "]" if (@option_nocurves)
	msg += " [" + @msg_CPoint + "]" if (@option_cpoint)
	
	Sketchup.set_status_text msg	
	Sketchup.set_status_text @msg_Distance, SB_VCB_LABEL
	Sketchup.set_status_text d.to_l, SB_VCB_VALUE
end

end	#End Class TOSToolLine
	
end	#End Module SUToolsOnSurface
